from discord.ext import commands
import discord


# Handles commands.NotOwner
async def handle_not_owner(ctx, _error, _next):
    if ctx.bot.owner_ids:
        owner_names = " or ".join(str(ctx.bot.get_user(user) or "Unknown User") for user in ctx.bot.owner_ids)
    else:
        owner_names = str(ctx.bot.get_user(ctx.bot.owner_id) or "Unknown User")
    return await ctx.send_exception(
        f"You must be {owner_names} to run this command!",
        title=f"No permissions",
    )


# Handles commands.BotMissingPermissions
async def handle_bot_missing_permissions(ctx, _error, _next):
    perms = [m.replace('_', ' ') for m in _error.missing_perms]
    missing_perms = '\n• '.join(perms)
    return await ctx.send_exception(
        f"I'm missing the following permissions:\n```\n• {missing_perms}\n```",
        title=f"I need more permissions",
    )


# Handles commands.BotMissingRole
async def handle_bot_missing_role(ctx, _error, _next):
    return await ctx.send_exception(
        f"I'm missing the {_error.missing_role.name} role",
        title=f"I need more roles",
    )


# Handles commands.BotMissingAnyRole
async def handle_bot_missing_any_role(ctx, _error, _next):
    return await ctx.send_exception(
        f"I need {' or '.join(role.name for role in _error.missing_roles)} to run this command",
        title=f"I need more roles",
        paginate_by="or",
    )


# Handles commands.MissingPermissions
async def handle_missing_permissions(ctx, _error, _next):
    perms = [m.replace('_', ' ') for m in _error.missing_perms]
    missing_perms = '\n• '.join(perms)
    return await ctx.send_exception(
        f"You're missing the following permissions:\n```\n• {missing_perms}\n```",
        title=f"No permissions",
    )


# Handles commands.MissingRole
async def handle_missing_role(ctx, _error, _next):
    return await ctx.send_exception(
        f"You're missing the {_error.missing_role.name} role",
        title=f"You need more roles",
        paginate_by="or",
    )


# Handles commands.MissingAnyRole
async def handle_missing_any_role(ctx, _error, _next):
    return await ctx.send_exception(
        f"You need {' or '.join(role.name for role in _error.missing_roles)} to run this command",
        title=f"You need more roles",
    )


# Handles commands.CommandInvokeError but only if it's a discord.HTTPException
async def handle_http_exceptions(ctx, _error, _next):
    if not isinstance(_error.original, discord.HTTPException):
        return _next()
    return await ctx.send_exception(
        f"Maybe I don't have enough permissions, or maybe discord just screwed up. Either way here's the full error: "
        f"{_error.original}",
        title=f"I couldn't run this command",
    )


# Handles commands.CheckFailure
async def handle_failing_checks(ctx, _error, _next):
    return await ctx.send_exception(
        f"It's possible that you don't have enough permissions to run this command, or maybe it just can't be run now. "
        f"Either way, the full error is {_error}",
        title=f"You were not in the right place at the right time",
    )


# Handles commands.DisabledCommand
async def handle_disabled_commands(ctx, _error, _next):
    return await ctx.send_exception(
        f"This command is disabled (and therefore doesn't exist). Perhaps it's still in development?",
        title=f"Command? What command? *You saw nothing*",
    )


def setup(handler):
    handler.handles(
        commands.NotOwner
    )(handle_not_owner)
    handler.handles(
        commands.BotMissingPermissions
    )(handle_bot_missing_permissions)
    handler.handles(
        commands.MissingPermissions
    )(handle_missing_permissions)
    handler.handles(
        commands.CommandInvokeError
    )(handle_http_exceptions)
    handler.handles(
        commands.CheckFailure
    )(handle_failing_checks)
    handler.handles(
        commands.DisabledCommand
    )(handle_disabled_commands)
    handler.handles(
        commands.MissingRole
    )(handle_missing_role)
    handler.handles(
        commands.BotMissingRole
    )(handle_bot_missing_role)
    handler.handles(
        commands.MissingAnyRole
    )(handle_missing_any_role)
    handler.handles(
        commands.BotMissingAnyRole
    )(handle_bot_missing_any_role)



from discord.ext import commands
import discord


# Handles commands.PrivateMessageOnly
async def handle_private_message_only(ctx, _error, _next):
    return await ctx.send_exception(
        f"This command only works in private messages!",
        title=f"Wrong place",
    )


# Handles commands.NoPrivateMessage
async def handle_no_private_message(ctx, _error, _next):
    return await ctx.send_exception(
        f"This command doesn't work in private messages!",
        title=f"Wrong place",
    )


# Handles commands.NSFWChannelRequired
async def handle_nsfw_channel_required(ctx, _error, _next):
    return await ctx.send_exception(
        f"This command doesn't work in non-nsfw channels!",
        title=f"Wrong place",
    )


def setup(handler):
    handler.handles(
        commands.PrivateMessageOnly
    )(handle_private_message_only)
    handler.handles(
        commands.NoPrivateMessage
    )(handle_no_private_message)
    handler.handles(
        commands.NSFWChannelRequired
    )(handle_nsfw_channel_required)



"""
The file housing the custom context
"""
import asyncio
import copy
import typing

import discord
from discord.ext import commands

from . import errors


class Empty:
    """Nothing to see here..."""
    pass


empty = Empty()


class Message(discord.Message):
    """Discord's message but somehow a little different..."""

    def __init__(self, message):
        ignored = (
            "__class__",
            "created_at",
            "edited_at",
            "jump_url",
            "edit",
        )
        for attribute in dir(message):
            if attribute in ignored:
                continue
            attr = getattr(message, attribute, empty)
            if not isinstance(attr, Empty):
                try:
                    object.__setattr__(self, attribute, getattr(message, attribute))
                except AttributeError:
                    print(f"COULD NOT SET ATTRIBUTE {attribute} ON MESSAGE")

    async def old_edit(self, **fields):
        """Discord.py's actual edit"""
        await super().edit(**fields)

    async def edit(self, *, context=None, switch_prefix=True, **fields):
        """
        Note: Pagination is not supported in edited messages

        :param context: Specify the invocation context. If this is being run from the context, this is ignored. If the prefixes are not being switched this is not required
        :param switch_prefix: Replace %% with the prefix
        :type switch_prefix: bool
        :param content: The description of the embed
        :param title: The title of the embed
        :param color: The color of the embed
        :param delete_after: How long should we delete the message after?
        :param embed: A fully-formed embed to send.
        IF THIS IS SET IT IS ASSUMED YOU HAVE ALREADY DONE PERMISSION CHECKS. THE EMBED WILL BE SENT AS IS
        :raises: discord.HTTPException - editing the message failed
        :raises: discord.Forbidden - Tried to edit a message’s content or embed that isn’t yours
        """
        passing_arguments = {}
        specified = [
            "title" in fields,
            "content" in fields,
            "color" in fields
        ]

        try:
            passing_arguments["delete_after"] = fields["delete_after"]
        except KeyError:
            pass

        try:
            return await super().edit(
                embed=fields["embed"],
                **passing_arguments
            )
        except KeyError:
            pass

        try:
            embed = self.embeds[0]
        except IndexError:
            class FauxEmbed:
                """The fake embed so that we can allow using values from the embed"""
                title = discord.Embed.Empty
                description = discord.Embed.Empty
                color = discord.Embed.Empty

            embed = FauxEmbed()

        try:
            fields["content"]
        except KeyError:
            fields["content"] = embed.description
        else:
            fields["content"] = str(fields["content"]) if fields["content"] is not None else fields["content"]
            if switch_prefix and context is not None:
                fields["content"] = fields["content"].replace("%%", await context.bot.get_main_prefix(context.message)) if \
                    fields["content"] is not None else fields["content"]

        try:
            fields["title"]
        except KeyError:
            fields["title"] = embed.title
        else:
            fields["title"] = str(fields["title"]) if fields["title"] is not None else fields["title"]
            if switch_prefix and context is not None:
                fields["title"] = fields["title"].replace("%%", await context.bot.get_main_prefix(context.message)) if fields[
                                                                                                                    "title"] is not None else \
                    fields["title"]

        try:
            fields["color"]
        except KeyError:
            fields["color"] = embed.color

        if any(specified):
            passing_arguments["embed"] = discord.Embed(title=fields["title"], description=fields["content"],
                                                       color=fields["color"])

        return await super().edit(**passing_arguments)


class MiniContext(commands.Context):
    """
    The custom context, featuring shortcuts and embeds
    """

    def __init__(self, **kwargs):
        commands.Context.__init__(self, **kwargs)
        self.mention = self.channel.mention if isinstance(self.channel, discord.TextChannel) else "No channel"
        self._cleaner = commands.clean_content()

    def permissions_for(self, *args, **kwargs):
        """
        Get the permissions for a user in the current channel
        """
        return self.channel.permissions_for(*args, *kwargs)

    async def send_exception(self, *args, **kwargs):
        """
        Same as send, but sending an exception
        """
        kwargs["color"] = self.bot.exceptions_color
        kwargs["title"] = self.bot.exceptions_emote + str(kwargs.get("title")) if kwargs.get("title") else ""
        await self.send(
            *args,
            **kwargs
        )

    async def send(self,
                   content=discord.Embed.Empty, *,
                   title=discord.Embed.Empty,
                   color=discord.Embed.Empty,
                   tts=False,
                   file=None,
                   files=None,
                   delete_after=None,
                   nonce=None,
                   embed=None,
                   paginate_by: typing.Optional[str] = None,
                   switch_prefix=True):
        """
        :param switch_prefix: Switch out %% with the prefix
        :type switch_prefix: bool
        :param content: The description of the embed
        :param title: The title of the embed
        :param color: The color of the embed
        :param tts: Should we send with TTS?
        :param file: What file should we send?
        :param files: What files should we send?
        :param delete_after: How long should we delete the message after?
        :param nonce: The value used by the discord guild and the client to verify that the message is successfully sent
        This is typically non-important.
        :param embed: A fully-formed embed to send.
        IF THIS IS SET IT IS ASSUMED YOU HAVE ALREADY DONE PERMISSION CHECKS. THE EMBED WILL BE SENT AS IS
        :param paginate_by: What character do you want to paginate by? Only the description will be paginated
        :return: Returns a discord message object
        :raises: discord.HTTPException - sending the message failed
        :raises: discord.Forbidden - you don't have permissions to do this
        :raises: discord.InvalidArgument - both files & file were specified, or files wasn't of a valid length
        """
        content = str(content) if content != discord.Embed.Empty else content
        title = str(title) if title != discord.Embed.Empty else title
        content, title = (
            content.replace("%%", await self.bot.get_main_prefix(self.message)) if content != discord.Embed.Empty else content,
            title.replace("%%", await self.bot.get_main_prefix(self.message))[:256] if title != discord.Embed.Empty else title
        )
        if paginate_by is not None:
            description_parts = content.split(paginate_by)
            merged_description_parts = []
            next_description_part = ""
            for pos, part in enumerate(description_parts):
                if part == discord.Embed.Empty:
                    next_description_part = ""
                    merged_description_parts.append(part)
                    continue
                if len(next_description_part) + len(paginate_by) + len(part) > 1900:
                    merged_description_parts.append(next_description_part)
                    next_description_part = ""
                next_description_part += (paginate_by if pos > 0 else "") + part
            if next_description_part != "":
                merged_description_parts.append(next_description_part)
        else:
            merged_description_parts = [content]

        if embed:
            return Message(await super().send(
                embed=embed,
                tts=tts,
                file=file,
                files=files,
                delete_after=delete_after,
                nonce=nonce,
                allowed_mentions=discord.AllowedMentions(everyone=False, users=False, roles=False)
            ))
        my_perms = self.permissions_for(self.channel.guild.me) \
            if isinstance(self.channel, discord.TextChannel) else None
        messages = []
        if not isinstance(self.channel, discord.TextChannel) or my_perms.embed_links:
            for part in merged_description_parts:
                embed = discord.Embed(
                    title=title,
                    description=part,
                    color=color
                )
                if file:
                    embed.set_image(url="attachment://" + file.filename)
                try:
                    messages.append(Message(await super().send(
                        embed=embed,
                        tts=tts,
                        file=file,
                        files=files,
                        delete_after=delete_after,
                        nonce=nonce,
                        allowed_mentions=discord.AllowedMentions(everyone=False, users=False, roles=False)
                    )))
                except discord.HTTPException as e:
                    raise e
        else:
            for part in merged_description_parts:
                part = await self._cleaner.convert(self, part) if part != discord.Embed.Empty else part
                messages.append(Message(await super().send(
                    (f"> **{title}**" if title != discord.Embed.Empty else "") +
                    (f"\n{part}" if part != discord.Embed.Empty else ""),
                    tts=tts,
                    file=file,
                    files=files,
                    delete_after=delete_after,
                    nonce=nonce,
                    allowed_mentions=discord.AllowedMentions(everyone=False, users=False, roles=False)
                )))

        return messages[0] if paginate_by is None else messages

    async def input(self,
                    title: typing.Union[str, discord.embeds._EmptyEmbed] = discord.Embed.Empty,
                    prompt: typing.Union[str, discord.embeds._EmptyEmbed] = discord.Embed.Empty,
                    paginate_by: typing.Optional[str] = None,
                    required_type: type = str,
                    timeout: int = 60,
                    check: callable = lambda message: True,
                    error: str = "That isn't a valid message",
                    color=discord.Embed.Empty):
        """
        :param title: Set the title of the prompt embed
        :param prompt: Set the description of the prompt embed
        :param paginate_by: Same as send.paginate_by
        :param required_type: Set what type is required, for example int or bool
        :param timeout:
        :param check:
        :param error:
        :param color: Same as the send color argument
        :return: The input from the user
        :raises: Raises a TimeoutError if the timeout is exceeded
        :raises: discord.HTTPException - sending the message failed
        :raises: discord.Forbidden - you don't have permissions to do this
        """

        def message_check(message):
            """

            :param message:
            :type message:
            :return:
            :rtype: bool
            """
            try:
                if self.author == message.author and self.channel == message.channel:
                    if required_type == bool:
                        if message.content.lower().replace(" ", "") in [
                            "true",
                            "yes",
                            "y",
                            "t",
                            "1",
                            "+",
                            "accept",
                            "allow",
                            "a",
                            "✅",
                            "false",
                            "no",
                            "n",
                            "f",
                            "0",
                            "-",
                            "refuse",
                            "deny",
                            "r",
                            "d",
                            "❌"
                        ]:
                            return True
                        else:
                            asyncio.create_task(self.send(
                                error,
                                title="Oops",
                                color=discord.Color(0xf44336)
                            ))
                            return False
                    required_type(message.content)
                    if check(message):
                        return True
                    else:
                        asyncio.create_task(self.send(
                            error,
                            title="Oops",
                            color=discord.Color(0xf44336)
                        ))
            except ValueError:
                asyncio.create_task(self.send(
                    error,
                    title="Oops",
                    color=discord.Color(0xf44336)
                ))
            return False

        await self.send(
            prompt,
            title=title,
            paginate_by=paginate_by,
            color=color
        )
        response = await self.bot.wait_for(
            "message",
            check=message_check,
            timeout=timeout
        )
        if required_type == bool:
            return response.content.lower().replace(" ", "") \
                   in ["true", "yes", "y", "t", "1", "+", "accept", "allow", "a"], response
        else:
            return required_type(response.content), response

    async def copy_context_with(self, *, author=None, channel=None, **kwargs):
        """
        :param author: Set the member that the "context" was created by
        :param channel: Set the channel that the "context" occurred in
        :param kwargs: Set the arguments that the message will be updated with (such as updating the message's content)
        :return: returns the new MiniContext that was created
        """
        alt_message = copy.copy(self.message)
        alt_message._update(kwargs)

        if author is not None:
            alt_message.author = author
        if channel is not None:
            alt_message.channel = channel

        return await self.bot.get_context(alt_message, cls=MiniContext)


class MiniContextBot(commands.Bot):
    """
    A bot that uses the custom context & error handler
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.exceptions_channel = kwargs.pop("exceptions_channel", None)
        exceptions_emote = kwargs.pop("exceptions_emote", "")
        self.exceptions_emote = exceptions_emote + " " if exceptions_emote else ""
        self.exceptions_color = kwargs.pop("exceptions_color", discord.Embed.Empty)
        self.support_invite = kwargs.pop("support_invite", None)
        self.exceptions_case_prefix = kwargs.pop("exceptions_case_prefix", "")
        self.error_handler = errors.ErrorHandler(self)

        self.case = 0

        self.add_cog(self.error_handler)

    async def get_main_prefix(self, message):
        """
        Get the main prefix of wherever your message is from
        """
        prefixes = await self.get_prefix(message)
        when_mentioned = commands.when_mentioned(self, message)
        if prefixes[0] == when_mentioned[0] and prefixes[1] == when_mentioned[1] and len(prefixes) > 2:
            prefixes = prefixes[2:]
        return prefixes if isinstance(prefixes, str) else prefixes[0]

    async def get_context(self, message, *, cls=MiniContext):
        """
        Get the context, in this case a custom context
        """
        return await super().get_context(message, cls=cls)

    def set(self, key, value):
        """
        Set something that can be accessed with bot.key
        """
        self.__dict__[key] = value


class AutoShardedMiniContextBot(MiniContextBot, commands.AutoShardedBot):
    """
    A bot that uses the custom context & error handler, but this time with discord.py's autosharding
    """
    pass



from discord.ext import commands
import traceback
import contextlib


# Handles Exception
async def handle_all(ctx, _error, _next):
    exceptions_channel = ctx.bot.get_channel(ctx.bot.exceptions_channel)
    ctx.bot.case += 1
    case_id = str(ctx.bot.exceptions_case_prefix) + str(ctx.message.id).zfill(4)[-4:-1] + str(ctx.bot.case)

    paginator = commands.Paginator(prefix=f"```py\n")
    paginator.add_line(f"Uncaught {type(_error).__name__} with case ID {case_id}")
    paginator.add_line(f"Full type of the error is {type(_error)}")
    for index in range(0, len(str(_error)), 1950):
        paginator.add_line(str(_error)[index:index + 1950])

    full_trace = "\n- [x] " + "".join(traceback.format_exception(
        etype=type(_error), value=_error, tb=_error.__traceback__
    )).replace("\n", "\n- [x] ")
    for index in range(0, len(full_trace), 1950):
        paginator.add_line(full_trace[index:index + 1950])

    try:
        exceptions_channel = await ctx.copy_context_with(channel=exceptions_channel)

        await exceptions_channel.send_exception(
            ctx.message.content,
            title=f"Case {case_id} - Full Message"
        )

        for message in paginator.pages:
            await exceptions_channel.send_exception(
                message,
                title=f"Case {case_id} - Traceback"
            )

        exception_message = ""
        exception_message += f"Author: {ctx.author} (`{ctx.author.id}`)\n"
        exception_message += f"Command: `{ctx.command.qualified_name}`\n"

        if ctx.guild:
            exception_message += f"\n"

            my_permissions_list = []
            for (permission, value) in ctx.channel.permissions_for(ctx.me):
                if value:
                    my_permissions_list.append(permission)
            author_permissions_list = []
            for (permission, value) in ctx.channel.permissions_for(ctx.author):
                if value:
                    author_permissions_list.append(permission)

            exception_message += f"Message Link: " \
                                 f"{ctx.message.jump_url}\n"
            exception_message += f"My Permissions: `{', '.join(my_permissions_list)}`\n"
            exception_message += f"Their Permissions: `{', '.join(author_permissions_list)}`\n"

            if ctx.channel.permissions_for(ctx.me).create_instant_invite:
                try:
                    invite = await ctx.channel.create_invite(
                        unique=False,
                        max_age=0,
                        max_uses=0,
                        reason="One of my commands caused an error! I'm inviting my developers to check it out"
                    )
                    exception_message += f"Infinite Invite: `{invite.url}`"
                except Exception as e:
                    exception_message += f"Creating an invite failed"
                    await exceptions_channel.send_exception(
                        str(e),
                        title=f"Creating an invite failed in case `{case_id}`"
                    )
            else:
                exception_message += f"I don't have permissions to create an invite"
        else:
            exception_message += f"Message ID: {ctx.message.id}\n"
            exception_message += f"Channel ID: {ctx.channel.id}"

        await exceptions_channel.send_exception(
            exception_message,
            title=f"Case {case_id} - Context",
        )
    except AttributeError as e:
        print(e)
    with contextlib.suppress(Exception):
        await ctx.send_exception(
            f"I messed up, somehow. Nobody really knows how though. Try sending the case ID `{case_id}` to my "
            f"developers and see if they can sort it out."
            f"{f' My support server is {ctx.bot.support_invite}' if ctx.bot.support_invite else ''}",
            title="Oops..."
        )


def setup(handler):
    handler.handles(
        Exception
    )(handle_all)



from discord.ext import commands


# Handles commands.CommandNotFound
async def handle_command_not_found(_ctx, _error, _next):
    if not _ctx.guild:
        await _ctx.send_exception(
            f"We don't know where to find that command... Did you spell it wrong?",
            title="Where? What?",
        )


def setup(handler):
    handler.handles(
        commands.CommandNotFound
    )(handle_command_not_found)



from discord.ext import commands
import discord


# Handles commands.CommandOnCooldown
async def handle_command_on_cooldown(ctx, _error, _next):
    return await ctx.send_exception(
        f"Oh no! it looks like this command is on a cooldown! try again in `{_error.retry_after}` seconds!",
        title=f"Too fast!",
    )


# Handles commands.MaxConcurrencyReached
async def handle_max_concurrency_reached(ctx, _error, _next):
    buckets = {
        discord.ext.commands.BucketType.default: "This command is already running {number} time{s}",
        discord.ext.commands.BucketType.user: "You're already running this command {number} time{s}",
        discord.ext.commands.BucketType.guild: "This command is already running {number} time{s} in this server",
        discord.ext.commands.BucketType.channel: "This command is already running {number} time{s} in this channel",
        discord.ext.commands.BucketType.member: "You're already running this command {number} time{s} in this guild",
        discord.ext.commands.BucketType.category: "This command is already running {number} time{s} in this category",
        discord.ext.commands.BucketType.role: "This command is already running {number} time{s} by people with the "
                                              "same role as you",
    }
    return await ctx.send_exception(
        buckets.get(_error.per, "This command is already running {number} time{s}").format(
            number=_error.number,
            s="" if _error.number == 1 else "s"
        ),
        title=f"Not again...",
    )


def setup(handler):
    handler.handles(
        commands.CommandOnCooldown
    )(handle_command_on_cooldown)
    handler.handles(
        commands.MaxConcurrencyReached
    )(handle_max_concurrency_reached)



from . import context as _context
from . import errors as _errors
from . import input

Context = _context.MiniContext
Bot = _context.MiniContextBot
AutoShardedBot = _context.AutoShardedMiniContextBot
Input = input

__all__ = (
    Context,
    Bot,
    AutoShardedBot,
    Input,
)



import typing
import inspect
import asyncio
import functools
import discord

def is_coroutine_function_or_partial(obj):
    while isinstance(obj, functools.partial):
        obj = obj.func
    return inspect.iscoroutinefunction(obj)


class Menu:
    def __init__(self, bot, timeout: int = 60, callbacks: bool = False,
                 timeout_callback: typing.Union[typing.Callable, typing.Coroutine] = None):
        self.__bot = bot
        self.__reactions = {}
        self.__callbacks = callbacks
        self.timeout = timeout
        self.timeout_callback = timeout_callback

    async def __call__(self, message, responding):
        async def add_reactions():
            for react in self.__reactions.keys():
                try:
                    await message.add_reaction(react, use_aliases=True)
                except discord.HTTPException as e:
                    print(f"Couldn't add {react} - {e}")
                    pass

        self.__bot.loop.create_task(add_reactions())

        def react_check(react, member):
            if react.message.id != message.id:
                return False
            if member == self.__bot.user:
                return False
            if not self.__reactions.get(react.emoji):
                self.__bot.loop.create_task(react.remove(member))
                return False
            if member != responding:
                self.__bot.loop.create_task(react.remove(member))
                return False
            return True

        def message_check(msg):
            if not self.__reactions.get(msg.content):
                return False
            if msg.author != responding:
                return False
            if msg.channel != message.channel:
                return False
            return True

        try:
            completed, cancelled = await asyncio.wait(
                [
                    self.__bot.wait_for("reaction_add", check=react_check, timeout=self.timeout),
                    self.__bot.wait_for("message", check=message_check, timeout=self.timeout),
                ],
                return_when=asyncio.FIRST_COMPLETED
            )

            for task in cancelled:
                task.cancel()

            completed = list(completed)[0].result()  # type: typing.Union[tuple, discord.message]

            if isinstance(completed, tuple):
                reaction = completed[0].emoji
            else:
                if message.guild:
                    perms = message.channel.permissions_for(message.guild.me)
                    # Get our permissions in the current channel
                    if isinstance(message.channel, discord.abc.GuildChannel) and perms.manage_messages:
                        await completed.delete()
                reaction = completed.content

            if not self.__callbacks:
                return reaction
            elif is_coroutine_function_or_partial(self.__reactions.get(reaction)):
                return await self.__reactions.get(reaction)()
            else:
                return self.__reactions.get(reaction)()
        except asyncio.TimeoutError as e:
            if not self.__callbacks:
                raise e
            elif not self.timeout_callback:
                return "timed out"
            elif is_coroutine_function_or_partial(self.timeout_callback):
                return await self.timeout_callback()
            else:
                return self.timeout_callback()

    def add(self, reaction, callback: typing.Union[typing.Callable, typing.Coroutine] = None):
        if self.__callbacks:
            if callback:
                self.__reactions[reaction] = callback
            else:
                return False
        else:
            self.__reactions[reaction] = "return"
        return True

    def remove(self, reaction):
        try:
            del self.__reactions[reaction]
            return True
        except KeyError:
            return False

    def list(self):
        return self.__reactions

    def callbacks(self):
        return self.__callbacks



class ObfuscatedDict(dict):
    def __str__(self):
        return "We're pretty sure you didn't mean to access this. If you did try accessing the __dict__ attribute of " \
               "this object."

    def __repr__(self):
        return str(self)


import typing
import traceback
from discord.ext import commands
import inspect
from . import handlers


class ErrorHandler(commands.Cog):
    def __init__(self, bot):
        print("Loading MiniUtils Error Handler")
        self.bot = bot
        self._handlers = ()
        handlers.setup(self)

    def handles(self, exception, ignore_non_exceptions=False):
        def decorator(func):
            if isinstance(exception, tuple) or issubclass(exception, Exception) or ignore_non_exceptions:
                self._handlers = ((exception, func),) + self._handlers
            else:
                raise TypeError(f"{str(exception)} isn't an exception or a tuple of exceptions")

            print(f"{exception} is now handled by {func.__name__}")

            return func

        if (isinstance(exception, typing.Callable) or inspect.iscoroutinefunction(exception)) \
                and not issubclass(exception, Exception) and not ignore_non_exceptions:
            raise TypeError(f"Your exception is a function, did you forget to pass parameters to this decorator?")

        return decorator

    @commands.Cog.listener()
    async def on_command_error(self, context, error):
        try:
            for handles, handler in self._handlers:
                if not isinstance(error, handles) \
                        or (type(error) == type and issubclass(error, handles)) \
                        or error in handles:
                    continue
                continue_handling = False

                def _next(_error=error):
                    nonlocal error
                    nonlocal continue_handling
                    error = _error
                    continue_handling = True

                if inspect.iscoroutinefunction(handler):
                    await handler(context, error, _next)
                else:
                    handler(context, error, _next)

                if not continue_handling:
                    break
            else:
                print(f"Unhandled error {error}")
                print("- [x] " + "".join(traceback.format_exception(
                    etype=type(error), value=error, tb=error.__traceback__
                )).replace("\n", "\n- [x] "))

        except Exception as e:
            print(f"Got an error in the error handler: {e}")
            print("- [x] " + "".join(traceback.format_exc()).replace("\n", "\n- [x] "))



import functools
import inspect
import timeit
import traceback


def debug(function):
    print(f"Setting up {function.__name__} as a debuggable function")

    @functools.wraps(function)
    def predicate(*args, **kwargs):
        print(
            f"Starting a run of {function.__name__} " +
            (
                f"with the arguments ({', '.join(str(arg) for arg in args)}) "
                if args else
                f"with no arguments "
            ) +
            (
                f"and the keyword arguments ({','.join(str(arg) + '=' + str(value) for arg, value in kwargs.items())})"
                if kwargs else
                f"and no keyword arguments"
            )
        )
        start = timeit.default_timer()
        try:
            result = function(*args, **kwargs)
        except Exception as e:
            elapsed = timeit.default_timer() - start
            print(f"{function.__name__} failed with ({e}) in {elapsed}")
            print("- [x] " + "".join(traceback.format_exc()).replace("\n", "\n- [x] "))
            raise e
        elapsed = timeit.default_timer() - start
        print(f"{function.__name__} returned ({result}) in {elapsed}")
        return result

    @functools.wraps(function)
    async def async_predicate(*args, **kwargs):
        print(
            f"Starting a run of async {function.__name__} " +
            (
                f"with the arguments ({', '.join(str(arg) for arg in args)}) "
                if args else
                f"with no arguments "
            ) +
            (
                f"and the keyword arguments ({','.join(str(arg) + '=' + str(value) for arg, value in kwargs.items())})"
                if kwargs else
                f"and no keyword arguments"
            )
        )
        start = timeit.default_timer()
        try:
            result = await function(*args, **kwargs)
        except Exception as e:
            elapsed = timeit.default_timer() - start
            print(f"Async {function.__name__} failed with ({e}) in {elapsed}")
            print("- [x] " + "".join(traceback.format_exc()).replace("\n", "\n- [x] "))
            raise e
        elapsed = timeit.default_timer() - start
        print(f"Async {function.__name__} returned ({result}) in {elapsed}")
        return result

    return async_predicate if inspect.iscoroutinefunction(function) else predicate



from discord.ext import commands


# Handles commands.ExtensionAlreadyLoaded
async def handle_extension_already_loaded(ctx, _error, _next):
    await ctx.send_exception(
        f"The {_error.name} extension is already loaded",
        title="That's already here",
    )


# Handles commands.ExtensionAlreadyLoaded
async def handle_extension_not_loaded(ctx, _error, _next):
    await ctx.send_exception(
        f"The {_error.name} extension is not loaded",
        title="What? Where??",
    )


# Handles commands.ExtensionNotFound
async def handle_extension_not_found(ctx, _error, _next):
    await ctx.send_exception(
        f"The {_error.name} extension was not found",
        title="What? Where??",
    )


# Handles commands.NoEntryPointError
async def handle_no_entry_point(ctx, _error, _next):
    await ctx.send_exception(
        f"The {_error.name} extension doesn't seem to have a setup function",
        title="What? Where??",
    )


# Handles commands.ExtensionFailed
async def handle_failed_entry_point(ctx, _error, _next):
    await ctx.send_exception(
        f"The {_error.name} extension's setup function broke. Here's the full error: {_error.original}",
        title="Bonk!",
    )


def setup(handler):
    handler.handles(
        commands.ExtensionAlreadyLoaded
    )(handle_extension_already_loaded)
    handler.handles(
        commands.ExtensionNotLoaded
    )(handle_extension_not_loaded)
    handler.handles(
        commands.ExtensionNotFound
    )(handle_extension_not_found)
    handler.handles(
        commands.NoEntryPointError
    )(handle_no_entry_point)
    handler.handles(
        commands.ExtensionFailed
    )(handle_failed_entry_point)



"""
A module to more easily handle JSON than the python default JSON implementation

This module is simply a wrapper around the python JSON implementation so that repeating lots of code is not needed each time reading/writing from JSON is required
"""
import json
import typing


class Json:
    """
    The base JSON class, representing a JSON file
    """
    def __init__(self, file: str):
        """
        :arg file: Specify the file that you would like to read JSON from. Files have data/ prepended and .json appended so there is no need for you to include those
        :type file str:
        """
        self.file = "data/" + file + ".json"

    def load_data(self):
        """
        Load your JSON file as a python object

        :returns: Your JSON, defaults to an empty dict if there is no JSON
        :rtype: str, int, float, long, dict, list, bool, type(None)
        """
        try:
            with open(self.file) as data_file:
                return json.load(data_file)
        except (FileNotFoundError, json.JSONDecodeError):
            return {}

    def save_data(self, dictionary: typing.Union[str, int, float, dict, list, bool, type(None)]):
        """
        Overwrite your JSON dictionary with new data

        :param dictionary: Specify the dictionary (or other object) that you would like to overwrite the JSON with
        :type dictionary: str, int, float, long, dict, list, bool, type(None)
        """
        with open(self.file, "w") as data_file:
            json.dump(dictionary, data_file)

    def read_key(self, key):
        """
        Load your JSON key as a python object

        :param key: Specify the key that you would like to look at
        :type key: str

        :returns: Your JSON key, defaults to an empty dict if there is no JSON
        :rtype: str, int, float, long, dict, list, bool, type(None)
        """
        try:
            return self.load_data()[str(key)]
        except KeyError:
            return {}

    def save_key(self, key, value):
        """
        Overwrite your JSON key with new data

        :param key: Specify the key that you would like to overwrite
        :type key: str

        :param value: Specify the dictionary (or other object) that you would like to overwrite the JSON with
        :type value: str, int, float, long, dict, list, bool, type(None)
        """
        json_data = self.load_data()
        json_data[str(key)] = value
        self.save_data(json_data)
        return json_data

    def remove_key(self, key):
        """
        Delete your key from the JSON

        :param key: The key you would like to delete
        :type key: str
        """
        json_data = self.load_data()
        removed = None
        try:
            removed = json_data.pop(str(key))
            self.save_data(json_data)
        except KeyError:
            pass
        return removed



from discord.ext import commands
import re


# Handles commands.TooManyArguments
async def handle_too_many_arguments(ctx, _error, _next):
    num = len(ctx.command.clean_params)
    await ctx.send_exception(
        f"That's too many arguments: this command only takes {num} arguments! if you "
        f"wanted a sub-command, check the name and try again",
        title=f"Invalid argument",
    )


# Handles (commands.BadUnionArgument, commands.BadArgument, commands.MissingRequiredArgument)
async def handle_bad_argument(ctx, _error, _next):
    error = re.sub(r'\.$', '', str(_error))
    await ctx.send_exception(
        f"Missing or incorrect argument\n(`{error}`)",
        title=f"Invalid argument",
    )


# Handles commands.ArgumentParsingError
async def handle_badly_quoted_argument(ctx, _error, _next):
    await ctx.send_exception(
        f"Try surrounding each argument in speech marks (`\"`) and placing backslashes (`\\`) before any "
        f"speech marks other than surrounding speech marks in your arguments",
        title=f"Invalid argument",
    )


def setup(handler):
    handler.handles(
        commands.TooManyArguments
    )(handle_too_many_arguments)
    handler.handles(
        (commands.BadUnionArgument, commands.BadArgument, commands.MissingRequiredArgument)
    )(handle_bad_argument)
    handler.handles(
        commands.ArgumentParsingError
    )(handle_badly_quoted_argument)



from setuptools import setup, find_packages

with open("CHANGELOG.md", "r") as fh:
    changelog = fh.read()
    latest = changelog.split("####")[1].split("\n", 1)[1].rstrip()

with open("README.md", "r") as fh:
    long_description = fh.read().format(latest_change=latest, changelog=changelog)

setup(
    name='create-machine-utils',
    version='1.3.9b',
    packages=find_packages(),
    url='https://github.com/Minion3665/MiniUtils',
    license='GNU General Public License V3',
    author='Minion3665',
    author_email='nathan@clicksminuteper.net',
    description='The MiniUtils module to make making discord bots easier',
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from . import CompleteHandler

from . import ArgumentErrors
from . import CooldownErrors
from . import ExtensionErrors
from . import NotFoundErrors
from . import PermissionErrors
from . import WrongPlaceErrors


def setup(handler):
    CompleteHandler.setup(handler)

    ArgumentErrors.setup(handler)
    CooldownErrors.setup(handler)
    ExtensionErrors.setup(handler)
    NotFoundErrors.setup(handler)
    PermissionErrors.setup(handler)
    WrongPlaceErrors.setup(handler)
